#include "stdafx.h"
#include <chrono>
#include <stdio.h>
#include <string.h>

//#define FULL_TESTING

#if defined(FULL_TESTING) && defined(WIN32)
#include <intrin.h>
#include <algorithm>
#include "ETWProviders\etwprof.h"
#include <thread>

void ETWAllocAndFreeTest(const int bufSize)
{
	const int kElements = bufSize / sizeof(int);
	Sleep(100);
	const int kIterations = 100;
	char buffer[100];
	sprintf_s(buffer, "%d allocations and frees with no use", kIterations);
	CETWScope outerTimer(buffer);
	for (int loop = 0; loop < kIterations; ++loop)
	{
		int* p = nullptr;
		{
			CETWScope timer("Allocating for no use");
			p = new int[kElements];
		}
		{
			CETWScope timer("Freeing unused");
			delete[] p;
		}
	}
}

void ETWAllocAndUseTest(const int bufSize, const bool reuse)
{
	const int kElements = bufSize / sizeof(int);
	Sleep(100);
	int* p = nullptr;
	if (reuse)
	{
		// Allocate and zero -- note the "()".
		p = new int[kElements]();
	}
	// Give things some time to quiet down.
	Sleep(50);
	for (int loop = 0; loop < 15; ++loop)
	{
		if (!reuse)
		{
			CETWScope timer("Allocating for use");
			p = new int[kElements];
		}
		{
			CETWScope timer(reuse ? "Writing to old memory" : "Writing to new memory");
			for (int i = 0; i < kElements; ++i)
			{
				p[i] = 31; // Meaningless number.
			}
		}
		int sum = 0;
		{
			CETWScope timer("Reading memory");
			for (int i = 0; i < kElements; ++i)
			{
				sum += p[i];
			}
		}
		if (!reuse)
		{
			CETWScope timer("Freeing after use");
			delete[] p;
		}
		// Use the 'sum' variable so the compiler doesn't optimize it away.
		ETWMarkPrintf("Done, sum is %d", sum);
		Sleep(30);
	}
	if (reuse)
	{
		delete[] p;
	}
}

volatile bool quit = false;

void busy()
{
	while (!quit)
		;
}
#endif // FULL_TESTING

class Timer
{
public:
	Timer()
	{
		start = clock.now();
	}
	// Returns the duration in seconds.
	double GetElapsed()
	{
		auto end = clock.now();
		auto duration = end - start;
		return std::chrono::duration_cast<std::chrono::nanoseconds>(duration).count() * 1.e-9;
	}
private:
	std::chrono::steady_clock::time_point start;
	std::chrono::steady_clock clock;

	Timer(const Timer&) = delete;
	Timer operator=(const Timer*) = delete;
};

void BusyWait(int ms)
{
	auto end = std::chrono::steady_clock::now() + std::chrono::milliseconds(ms);
	for (;;)
	{
		if (std::chrono::steady_clock::now() > end)
			break;
	}
}

void FastMeasure()
{
	printf("Busy waiting to raise the CPU frequency...\n");
	// Busy wait for a second so that the CPUs ramp up to full speed.
	BusyWait(500);
	const int bufSize = 32 * 1024 * 1024;
	const int iterationCount = 100;
	{
		Timer timer;
		for (int i = 0; i < iterationCount; ++i)
		{
			int* p = new int[bufSize / sizeof(int)];
			delete[] p;
		}
		printf("%1.4f s to allocate %d MB %d times.\n", timer.GetElapsed(), bufSize / (1024 * 1024), iterationCount);
	}
	{
		Timer timer;
		double deleteTime = 0.0;
		for (int i = 0; i < iterationCount; ++i)
		{
			int* p = new int[bufSize / sizeof(int)];
			Timer deleteTimer;
			delete[] p;
			deleteTime += deleteTimer.GetElapsed();
		}
		printf("%1.4f s to allocate %d MB %d times (%1.4f s to delete).\n", timer.GetElapsed(), bufSize / (1024 * 1024), iterationCount, deleteTime);
	}
	{
		// Initialize and zero the memory.
		int* p = new int[bufSize / sizeof(int)]();
		{
			// Repeatedly write to the already allocated memory.
			Timer timer;
			for (int i = 0; i < iterationCount; ++i)
			{
				memset(p, 1, bufSize);
			}
			printf("%1.4f s to write %d MB %d times.\n", timer.GetElapsed(), bufSize / (1024 * 1024), iterationCount);
		}
		{
			// Repeatedly read from the already allocated memory.
			Timer timer;
			int sum = 0;
			for (int i = 0; i < iterationCount; ++i)
			{
				for (size_t index = 0; index < bufSize / sizeof(int); ++index)
				{
					sum += p[index];
				}
			}
			printf("%1.4f s to read %d MB %d times, sum = %d.\n", timer.GetElapsed(), bufSize / (1024 * 1024), iterationCount, sum);
		}
		delete[] p;
	}
	{
		// Repeatedly allocate, write, and free memory.
		Timer timer;
		double deleteTime = 0.0;
		for (int i = 0; i < iterationCount; ++i)
		{
			int* p = new int[bufSize / sizeof(int)];
			memset(p, 1, bufSize);
			Timer deleteTimer;
			delete[] p;
			deleteTime += deleteTimer.GetElapsed();
		}
		printf("%1.4f s to allocate and write %d MB %d times (%1.4f s to delete).\n", timer.GetElapsed(), bufSize / (1024 * 1024), iterationCount, deleteTime);
	}
	{
		// Repeatedly allocate, read, and free memory.
		Timer timer;
		int sum = 0;
		for (int i = 0; i < iterationCount; ++i)
		{
			int* p = new int[bufSize / sizeof(int)];
			for (size_t index = 0; index < bufSize / sizeof(int); ++index)
			{
				sum += p[index];
			}
			delete[] p;
		}
		printf("%1.4f s to allocate and read %d MB %d times, sum = %d.\n", timer.GetElapsed(), bufSize / (1024 * 1024), iterationCount, sum);
	}
}

int main(int argc, char* argv[])
{
	FastMeasure();

#ifdef FULL_TESTING
	printf("Doing slow testing for easier ETW analysis. Record a trace to see what is happening.\n");

	// Start up a busy thread and wait for the CPU frequency
	// to be raised.
	std::thread background = std::thread(busy);
	Sleep(50);

	// My last-level cache is 6 MB so start with smaller than
	// that and work up to quite large.
	for (int bufSize = 4 * 1024 * 1024; bufSize <= 32 * 1024 * 1024; bufSize += bufSize)
	{
		ETWMarkPrintf("Running tests with a %d MB buffer", bufSize / (1024 * 1024));
		ETWAllocAndFreeTest(bufSize);
		ETWAllocAndUseTest(bufSize, false);
		ETWAllocAndUseTest(bufSize, true);
	}
	quit = true;
	background.join();
	Sleep(50);
#endif // FULL_TESTING

	return 0;
}
